/*
 * A simple flash mapping code for BCM963xx board flash memory
 * It is simple because it only treats all the flash memory as ROM
 * It is used with chips/map_rom.c
 *
 *  Song Wang (songw@broadcom.com)
 */

#include <linux/module.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <asm/io.h>
#include <linux/mtd/mtd.h>
#include <linux/mtd/map.h>
//#include <linux/config.h>

/* 2006.11.22 patch */
#include <linux/mtd/partitions.h>

#include <board.h>
#include <bcmTag.h>
#include <bcm_map_part.h>
#define  VERSION	"1.0"

extern PFILE_TAG kerSysImageTagGet(void);

static struct mtd_info *mymtd;

/* 2006.11.22 patch */
#ifdef CONFIG_MTD_FLASH_2M
static struct mtd_partition brcm_partition_info[5];//follow JIM's way for POT
#else
static struct mtd_partition brcm_partition_info[6];//follow JIM's way for POT
#endif
#ifdef CONFIG_MTD_FLASH_8M
#define CFE_ADDR        0xbf000000
#else
#define CFE_ADDR        0xbfc00000
#endif
#define CFE_SIZE        (64 << 10)
#define NVRAM_SIZE      (64 << 10)
#define DPF_SIZE        (64 << 10)
#define TAG_SIZE       0x100

#ifdef CONFIG_MTD_FLASH_2M
#define FLASH_SIZE              (2048 << 10)
#endif
#ifdef CONFIG_MTD_FLASH_4M
#define FLASH_SIZE              (4096 << 10)
#endif
#ifdef CONFIG_MTD_FLASH_8M
#define FLASH_SIZE              (8192 << 10)
#endif

#define FS_KERNEL_SIZE  (FLASH_SIZE - FS_KERNEL_ADDR)

#define DPF_ADDR        CFE_SIZE
#define NVRAM_ADDR      (DPF_ADDR + DPF_SIZE)
#define FS_KERNEL_ADDR  (NVRAM_ADDR + NVRAM_SIZE)

/* end */

static map_word brcm_physmap_read16(struct map_info *map, unsigned long ofs)
{
	map_word val;
	
#if 0 /* We added voluntary preemption and locks to flash driver, so everything should go to flash driver such that locks are in effect */
    /* If the requested flash address is in a memory mapped range, use
     * __raw_readw.  Otherwise, use kerSysReadFromFlash.
     */
    if(((map->map_priv_1 & ~0xfff00000) + ofs + sizeof(short)) < map->map_priv_2)
	    val.x[0] = __raw_readw(map->map_priv_1 + ofs);
    else
#endif        
//        kerSysReadFromFlash( &val.x[0], map->map_priv_1 + ofs, sizeof(short) );
	val.x[0] = __raw_readw(map->map_priv_1 + ofs);
	
	return val;
}

static void brcm_physmap_copy_from(struct map_info *map, void *to, unsigned long from, ssize_t len)
{
#if 0 /* We added voluntary preemption and locks to flash driver, so everything should go to flash driver such that locks are in effect */
    /* If the requested flash address is in a memory mapped range, use
     * memcpy_fromio.  Otherwise, use kerSysReadFromFlash.
     */
    if( ((map->map_priv_1 & ~0xfff00000) + from + len) < map->map_priv_2 )
	    memcpy_fromio(to, map->map_priv_1 + from, len);
    else
#endif        
//        kerSysReadFromFlash( to, map->map_priv_1 + from, len );
	 memcpy_fromio(to, map->map_priv_1 + from, len);	
}

/* 2006.11.22 patch */
static void brcm_physmap_write16(struct map_info *map, __u16 d, unsigned long adr)
{
       __raw_writew(d, map->map_priv_1 + adr);
       mb();
}
void brcm_physmap_copy_to(struct map_info *map, unsigned long to, const void *from, ssize_t len)
{
       memcpy_toio(map->map_priv_1 + to, from, len);
}
/* end */

static struct map_info brcm_physmap_map = {
	.name		= "Physically mapped flash",
	.bankwidth	= 2,
	.read		= brcm_physmap_read16,
	.copy_from	= brcm_physmap_copy_from,
/* 2006.11.22 patch */
        .write = brcm_physmap_write16,
        .copy_to = brcm_physmap_copy_to
};

static int __init init_brcm_physmap(void)
{
        PFILE_TAG pTag = NULL;
//        u_int32_t rootfs_addr, kernel_addr;
	/* 2006.11.22 patch */
        u_int32_t rootfs_addr, kernel_addr, fs_len, cfe_len;

	printk("bcm963xx_mtd driver v%s\n", VERSION);

        /* Read the flash memory map from flash memory. */
        if (!(pTag = kerSysImageTagGet())) {
                printk("Failed to read image tag from flash\n");
                return -EIO;
        }

        rootfs_addr = (u_int32_t) simple_strtoul(pTag->rootfsAddress, NULL, 10) + BOOT_OFFSET;
        kernel_addr = (u_int32_t) simple_strtoul(pTag->kernelAddress, NULL, 10) + BOOT_OFFSET;
	
	/* 2006.11.22 patch */
       //brcm_physmap_map.size = kernel_addr - rootfs_addr;
       brcm_physmap_map.size = FLASH_SIZE;
       fs_len = kernel_addr - rootfs_addr;
       cfe_len=CFE_SIZE;
       /* Ron mapping from fs */
       //brcm_physmap_map.map_priv_1 = (unsigned long)rootfs_addr;
	brcm_physmap_map.map_priv_1 = (unsigned long)CFE_ADDR;
	printk("kernel_addr == 0x%x rootfs_addr == 0x%x\n",kernel_addr,rootfs_addr);


	/* Set map_priv_2 to the amount of flash memory that is memory mapped to
	 * the flash base address.  On the BCM6338, serial flash parts are only
	 * memory mapped up to 1MB even though the flash part may be bigger.
	 */
	brcm_physmap_map.map_priv_2 =(unsigned long)kerSysMemoryMappedFlashSizeGet();

	if (!brcm_physmap_map.map_priv_1) {
		printk("Wrong rootfs starting address\n");
		return -EIO;
	}
	
	if (brcm_physmap_map.size <= 0) {
		printk("Wrong rootfs size\n");
		return -EIO;
	}	
	/* 2006.11.22 patch */	
/*	mymtd = do_map_probe("map_rom", &brcm_physmap_map);
	if (mymtd) {
		mymtd->owner = THIS_MODULE;
		add_mtd_device(mymtd);

		return 0;
	}

	return -ENXIO;*/

        mymtd = do_map_probe("cfi_probe", &brcm_physmap_map);
        if (!mymtd)
                return -EIO;
        mymtd->owner = THIS_MODULE;
        /* Ron file system */
        brcm_partition_info[0].name = "fs";
        //brcm_partition_info[0].offset = (cfe_len + TAG_SIZE);
        brcm_partition_info[0].offset = (FS_KERNEL_ADDR + TAG_SIZE);
        brcm_partition_info[0].size = fs_len;
        brcm_partition_info[0].mask_flags = 0;

        /* Ron tag + file system + kernel */
        brcm_partition_info[1].name = "tag+fs+kernel";
        //brcm_partition_info[1].offset = cfe_len;
        brcm_partition_info[1].offset = FS_KERNEL_ADDR;
        brcm_partition_info[1].size = FS_KERNEL_SIZE;
        brcm_partition_info[1].mask_flags = 0;

        /* Ron bootloader */
        brcm_partition_info[2].name = "bootloader";
        brcm_partition_info[2].offset = 0x00;
        brcm_partition_info[2].size = cfe_len;
        brcm_partition_info[2].mask_flags = 0;

        /* Ron nvram */
        brcm_partition_info[3].name = "nvram";
        brcm_partition_info[3].offset = NVRAM_ADDR;
        brcm_partition_info[3].size = NVRAM_SIZE;
        brcm_partition_info[3].mask_flags = 0;

        /* JIM add for mapping to bootloader */
        brcm_partition_info[4].name = "bootloader";
        brcm_partition_info[4].offset = 0x00;
        brcm_partition_info[4].size = cfe_len;
        brcm_partition_info[4].mask_flags = 0;

        /* Allen: for DPF */
        brcm_partition_info[5].name = "DPF_file";
        brcm_partition_info[5].offset = DPF_ADDR;
        brcm_partition_info[5].size =    DPF_SIZE;
        brcm_partition_info[5].mask_flags = 0;

	add_mtd_partitions(mymtd, brcm_partition_info, 6);//follow JIM's way for POT
	return 0;
}

static void __exit cleanup_brcm_physmap(void)
{
	if (mymtd) {
		/* 2006.11.22 patch */
		del_mtd_partitions(mymtd);
		del_mtd_device(mymtd);
		map_destroy(mymtd);
	}
	if (brcm_physmap_map.map_priv_1) {
		brcm_physmap_map.map_priv_1 = 0;
	}
}

module_init(init_brcm_physmap);
module_exit(cleanup_brcm_physmap);


MODULE_LICENSE("GPL");
MODULE_AUTHOR("Song Wang songw@broadcom.com");
MODULE_DESCRIPTION("Configurable MTD map driver for read-only root file system");
